from typing import Literal, List, Optional
from .hotels_impl import HotelData, Guests, THSData

def create_filter(
    *,
    hotel_data: List[HotelData],
    guests: Guests,
    room_type: Optional[str] = None,
    amenities: Optional[List[str]] = None,
) -> THSData:
    """Create a filter for Google Hotels API. (``?ths=``)

    Args:
        hotel_data (list[HotelData]): Hotel data as a list.
        guests (Guests): Guests.
        room_type (str, optional): Room type. If not provided, taken from first HotelData.
        amenities (list[str], optional): Preferred amenities. If not provided, taken from first HotelData.
    """
    # Extract from first HotelData if not provided
    if not room_type and hotel_data and hasattr(hotel_data[0], "room_type"):
        room_type = hotel_data[0].room_type
    if amenities is None and hotel_data and hasattr(hotel_data[0], "amenities"):
        amenities = hotel_data[0].amenities
    return THSData.from_interface(
        hotel_data=hotel_data, 
        guests=guests, 
        room_type=room_type,
        amenities=amenities
    ) 


from fast_hotels.hotels_impl import HotelData, Guests
from fast_hotels import get_hotels

def test_get_hotels_returns_results():
    hotel_data = [HotelData(
        checkin_date="2025-06-23", 
        checkout_date="2025-06-25", 
        location="Tokyo",
        room_type="standard",
        amenities=["wifi", "breakfast"]
    )]
    guests = Guests(adults=2, children=1, infants=0)
    
    result = get_hotels(
        hotel_data=hotel_data,
        guests=guests,
        room_type="standard",
        amenities=["wifi", "breakfast"],
        fetch_mode="common",
        limit=5,
        sort_by="price"
    )
    
    assert len(result.hotels) > 0
    assert result.lowest_price is not None
    assert result.current_price is not None
    
    # Test that hotels have the expected structure
    for hotel in result.hotels:
        assert hotel.name is not None
        assert hotel.price is not None
        assert isinstance(hotel.amenities, list) 


"""Typed implementation for hotels API requests"""

import base64
from typing import Any, List, Optional, TYPE_CHECKING, Literal, Union
from dataclasses import dataclass

# Import the generated protobuf
try:
    from . import hotels_pb2 as PB
    PB_AVAILABLE = True
except ImportError:
    # Fallback for development
    PB = None
    PB_AVAILABLE = False

if TYPE_CHECKING:
    # Type hints for protobuf
    class PB:
        class Info:
            pass
        class HotelData:
            pass
        class Guests:
            pass
        class RoomType:
            pass
       

class HotelData:
    """Represents hotel search data.

    Args:
        checkin_date (str): Check-in date in YYYY-MM-DD format.
        checkout_date (str): Check-out date in YYYY-MM-DD format.
        location (str): Location (city name or IATA airport code).
        room_type (str, optional): Room type preference. Default is None.
        amenities (List[str], optional): Preferred amenities. Default is None.
    """

    __slots__ = ("checkin_date", "checkout_date", "location", "room_type", "amenities")
    checkin_date: str
    checkout_date: str
    location: str
    room_type: Optional[str]
    amenities: Optional[List[str]]

    def __init__(
        self,
        *,
        checkin_date: str,
        checkout_date: str,
        location: str,
        room_type: Optional[str] = None,
        amenities: Optional[List[str]] = None,
    ):
        self.checkin_date = checkin_date
        self.checkout_date = checkout_date
        self.location = location
        self.room_type = room_type
        self.amenities = amenities

    def attach(self, info: Any) -> None:  # type: ignore
        """Attach this hotel data to a protobuf Info object"""
        if not PB_AVAILABLE:
            return  # Skip if protobuf not available
        
        data = info.data.add()
        data.checkin_date = self.checkin_date
        data.checkout_date = self.checkout_date
        
        # Set location
        data.location.city = self.location
        data.location.country = ""  # Could be enhanced with country detection
        
        if self.room_type:
            data.room_type = self.room_type
            
        if self.amenities:
            data.amenities.extend(self.amenities)

    def __repr__(self) -> str:
        return (
            f"HotelData(checkin_date={self.checkin_date!r}, "
            f"checkout_date={self.checkout_date!r}, "
            f"location={self.location!r}, "
            f"room_type={self.room_type!r}, "
            f"amenities={self.amenities!r})"
        )


class Guests:
    """Represents guest information for hotel booking."""
    
    def __init__(
        self,
        *,
        adults: int = 1,
        children: int = 0,
        infants: int = 0,
    ):
        assert adults >= 1, "At least one adult is required"
        assert adults + children + infants <= 9, "Too many guests (> 9)"
        assert infants <= adults, "You must have at least one adult per infant"

        self.adults = adults
        self.children = children
        self.infants = infants

    def attach(self, info: Any) -> None:  # type: ignore
        """Attach guest information to a protobuf Info object"""
        if not PB_AVAILABLE:
            return  # Skip if protobuf not available
        
        # Add adults
        for _ in range(self.adults):
            info.guests.guests.append(PB.GuestType.ADULT)
        
        # Add children
        for _ in range(self.children):
            info.guests.guests.append(PB.GuestType.CHILD)
        
        # Add infants
        for _ in range(self.infants):
            info.guests.guests.append(PB.GuestType.INFANT)

    def __repr__(self) -> str:
        return f"Guests(adults={self.adults}, children={self.children}, infants={self.infants})"


class THSData:
    """``?ths=`` data for Google Hotels API. (internal)

    Use `THSData.from_interface` instead.
    """

    def __init__(
        self,
        *,
        hotel_data: List[HotelData],
        room_type: Any,  # type: ignore
        guests: Guests,
        amenities: Optional[List[str]] = None,
    ):
        self.hotel_data = hotel_data
        self.room_type = room_type
        self.guests = guests
        self.amenities = amenities

    def pb(self) -> Any:  # type: ignore
        """Create a protobuf Info object"""
        if not PB_AVAILABLE:
            # Return stub data if protobuf not available
            class StubInfo:
                def SerializeToString(self):
                    return b"stub_data"
            return StubInfo()
        
        info = PB.Info()
        info.room_type = self.room_type

        # Attach guests
        self.guests.attach(info)

        # Attach hotel data
        for hotel in self.hotel_data:
            hotel.attach(info)

        # Attach global amenities
        if self.amenities:
            info.amenities.extend(self.amenities)

        return info

    def to_string(self) -> bytes:
        return self.pb().SerializeToString()

    def as_b64(self) -> bytes:
        return base64.b64encode(self.to_string())

    @staticmethod
    def from_interface(
        *,
        hotel_data: List[HotelData],
        guests: Guests,
        room_type: str,
        amenities: Optional[List[str]] = None,
    ):
        """Use ``?ths=`` from an interface.

        Args:
            hotel_data (list[HotelData]): Hotel data as a list.
            guests (Guests): Guests.
            room_type (str): Room type.
            amenities (list[str], optional): Preferred amenities.
        """
        if not PB_AVAILABLE:
            room_t = 1  # STANDARD
        else:
            room_t = {
                "standard": PB.RoomType.STANDARD,
                "deluxe": PB.RoomType.DELUXE,
                "suite": PB.RoomType.SUITE,
            }[room_type]

        return THSData(
            hotel_data=hotel_data,
            room_type=room_t,
            guests=guests,
            amenities=amenities
        )

    def __repr__(self) -> str:
        return f"THSData(hotel_data={self.hotel_data!r}, amenities={self.amenities!r})" 


from __future__ import annotations

from dataclasses import dataclass
from typing import List, Literal, Optional


@dataclass
class Result:
    hotels: List[Hotel]
    lowest_price: Optional[float] = None
    current_price: Optional[float] = None


@dataclass
class Hotel:
    name: str
    price: float
    rating: Optional[float] = None
    url: Optional[str] = None
    amenities: List[str] = None
    
    def __post_init__(self):
        if self.amenities is None:
            self.amenities = [] 


from typing import Any
import asyncio
from playwright.async_api import async_playwright

async def fetch_with_playwright(url: str) -> str:
    """Local Playwright fallback for hotel scraping"""
    async with async_playwright() as p:
        browser = await p.chromium.launch()
        page = await browser.new_page()
        await page.goto(url)
        if page.url.startswith("https://consent.google.com"):
            await page.click('text="Accept all"')
        locator = page.locator('div.x2A2jf, div.GIPbOc.sSHqwe')
        await locator.wait_for()
        body = await page.evaluate(
            "() => document.querySelector('[role=\"main\"]').innerHTML"
        )
        await browser.close()
    return body

def local_playwright_fetch(params: dict) -> Any:
    """Local Playwright fallback function"""
    url = "https://www.google.com/travel/hotels?" + "&".join(f"{k}={v}" for k, v in params.items())
    body = asyncio.run(fetch_with_playwright(url))

    class DummyResponse:
        status_code = 200
        text = body
        text_markdown = body

    return DummyResponse 


from typing import Any

from .primp import Client

CODE = """\
import asyncio
import sys
from playwright.async_api import async_playwright

async def main():
    async with async_playwright() as p:
        browser = await p.chromium.launch()
        page = await browser.new_page()
        await page.goto("%s")
        locator = page.locator('div.x2A2jf, div.GIPbOc.sSHqwe')
        await locator.wait_for()
        body = await page.evaluate(
            \"\"\"() => {
                return document.querySelector('[role="main"]').innerHTML
            }\"\"\"
        )
        await browser.close()
    sys.stdout.write(body)

asyncio.run(main())
"""


def fallback_playwright_fetch(params: dict) -> Any:
    """Fallback to serverless Playwright when direct API fails"""
    client = Client(impersonate="chrome_100", verify=False)

    res = client.post(
        "https://try.playwright.tech/service/control/run",
        json={
            "code": CODE
            % (
                "https://www.google.com/travel/hotels"
                + "?"
                + "&".join(f"{k}={v}" for k, v in params.items())
            ),
            "language": "python",
        },
    )
    assert res.status_code == 200, f"{res.status_code} Result: {res.text_markdown}"
    import json

    class DummyResponse:
        status_code = 200
        text = json.loads(res.text)["output"]
        text_markdown = text

    return DummyResponse 


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: hotels.proto
# Protobuf Python Version: 5.29.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    3,
    '',
    'hotels.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0chotels.proto\x12\x0b\x66\x61st_hotels\"?\n\x08Location\x12\x0c\n\x04\x63ity\x18\x01 \x01(\t\x12\x0f\n\x07\x63ountry\x18\x02 \x01(\t\x12\x14\n\x0c\x61irport_code\x18\x03 \x01(\t\"\x9a\x01\n\tHotelData\x12\x14\n\x0c\x63heckin_date\x18\x01 \x01(\t\x12\x15\n\rcheckout_date\x18\x02 \x01(\t\x12\'\n\x08location\x18\x03 \x01(\x0b\x32\x15.fast_hotels.Location\x12\x16\n\troom_type\x18\x04 \x01(\tH\x00\x88\x01\x01\x12\x11\n\tamenities\x18\x05 \x03(\tB\x0c\n\n_room_type\"0\n\x06Guests\x12&\n\x06guests\x18\x01 \x03(\x0e\x32\x16.fast_hotels.GuestType\"\x8e\x01\n\x04Info\x12$\n\x04\x64\x61ta\x18\x01 \x03(\x0b\x32\x16.fast_hotels.HotelData\x12(\n\troom_type\x18\x02 \x01(\x0e\x32\x15.fast_hotels.RoomType\x12#\n\x06guests\x18\x03 \x01(\x0b\x32\x13.fast_hotels.Guests\x12\x11\n\tamenities\x18\x05 \x03(\t*P\n\x08RoomType\x12\x10\n\x0cUNKNOWN_ROOM\x10\x00\x12\x0c\n\x08STANDARD\x10\x01\x12\n\n\x06\x44\x45LUXE\x10\x02\x12\t\n\x05SUITE\x10\x03\x12\r\n\tEXECUTIVE\x10\x04*@\n\tGuestType\x12\x11\n\rUNKNOWN_GUEST\x10\x00\x12\t\n\x05\x41\x44ULT\x10\x01\x12\t\n\x05\x43HILD\x10\x02\x12\n\n\x06INFANT\x10\x03\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'hotels_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_ROOMTYPE']._serialized_start=446
  _globals['_ROOMTYPE']._serialized_end=526
  _globals['_GUESTTYPE']._serialized_start=528
  _globals['_GUESTTYPE']._serialized_end=592
  _globals['_LOCATION']._serialized_start=29
  _globals['_LOCATION']._serialized_end=92
  _globals['_HOTELDATA']._serialized_start=95
  _globals['_HOTELDATA']._serialized_end=249
  _globals['_GUESTS']._serialized_start=251
  _globals['_GUESTS']._serialized_end=299
  _globals['_INFO']._serialized_start=302
  _globals['_INFO']._serialized_end=444
# @@protoc_insertion_point(module_scope)



import csv
import logging
import requests
from functools import lru_cache
from io import StringIO

CSV_URL = "https://raw.githubusercontent.com/lxndrblz/Airports/refs/heads/main/airports.csv"

@lru_cache(maxsize=1)
def load_global_iata_map() -> dict:
    mapping = {}
    try:
        response = requests.get(CSV_URL, timeout=10)
        response.raise_for_status()
        f = StringIO(response.text)
        reader = csv.DictReader(f)
        for row in reader:
            iata = row.get("code", "").strip()
            city = row.get("city", "").strip()
            name = row.get("name", "").strip()
            country = row.get("country", "").strip()
            if iata:
                if city:
                    mapping[iata.upper()] = city
                elif name:
                    mapping[iata.upper()] = name
                elif country:
                    mapping[iata.upper()] = country
    except Exception as e:
        logging.error(f"Failed to fetch IATA mapping: {e}")
    return mapping

def get_city_from_iata(iata_code: str) -> str:
    """
    Return the city, or if not available, the airport name, or if not available, the country for the given IATA code.
    Falls back to the original string if not found.
    """
    mapping = load_global_iata_map()
    return mapping.get(iata_code.strip().upper(), iata_code)



from typing import List, Literal, Optional

from selectolax.lexbor import LexborHTMLParser, LexborNode

from .schema import Hotel, Result
from .hotels_impl import HotelData, Guests
from .filter import THSData
from .fallback_playwright import fallback_playwright_fetch
from .primp import Client, Response
from .utils import get_city_from_iata


def fetch(params: dict, location: str) -> Response:
    """Fast HTTP request to Google Hotels API"""
    client = Client(impersonate="chrome_126", verify=False)
    if not location:
        raise ValueError("No location provided for hotel search.")
    # Convert airport code to city name if needed
    city = get_city_from_iata(location)
    # Clean up city for URL
    location_url = city.strip().replace(' ', '+').lower()
    url = f"https://www.google.com/travel/hotels/{location_url}"
    res = client.get(url, params=params)
    assert res.status_code == 200, f"{res.status_code} Result: {res.text_markdown}"
    return res


def get_hotels_from_filter(
    filter: THSData,
    currency: str = "",
    *,
    mode: Literal["common", "fallback", "force-fallback", "local"] = "common",
    sort_by: Optional[str] = None,
    limit: Optional[int] = None,
) -> Result:
    """Get hotels using a filter with multiple fallback strategies"""
    data = filter.as_b64()
    params = {
        "ths": data.decode("utf-8"),
        "hl": "en",
        "curr": currency,
    }
    # Extract location from the first HotelData
    if not filter.hotel_data or not getattr(filter.hotel_data[0], 'location', None):
        raise ValueError("No location found in hotel filter. Please specify a valid location in HotelData.")
    location = filter.hotel_data[0].location
    if mode in {"common", "fallback"}:
        try:
            res = fetch(params, location)
        except AssertionError as e:
            if mode == "fallback":
                res = fallback_playwright_fetch(params)
            else:
                raise e
    elif mode == "local":
        from .local_playwright import local_playwright_fetch
        res = local_playwright_fetch(params)
    else:
        res = fallback_playwright_fetch(params)
    try:
        return parse_response(res, sort_by=sort_by, limit=limit)
    except RuntimeError as e:
        if mode == "fallback":
            return get_hotels_from_filter(filter, mode="force-fallback", sort_by=sort_by, limit=limit)
        raise e


def get_hotels(
    *,
    hotel_data: List[HotelData],
   
    guests: Guests,
    room_type: Literal["standard", "deluxe", "suite"] = "standard",
    fetch_mode: Literal["common", "fallback", "force-fallback", "local"] = "common",
    amenities: Optional[List[str]] = None,
    limit: Optional[int] = None,
    sort_by: Optional[str] = None,
) -> Result:
    """Main API function for getting hotels"""
    return get_hotels_from_filter(
        THSData.from_interface(
            hotel_data=hotel_data,
            guests=guests,
            room_type=room_type,
            amenities=amenities,
        ),
        mode=fetch_mode,
        sort_by=sort_by,
        limit=limit,
    )


def parse_response(
    r: Response, *, dangerously_allow_looping_last_item: bool = False, sort_by: Optional[str] = None, limit: Optional[int] = None
) -> Result:
    """Parse the HTML response from Google Hotels"""
    class _blank:
        def text(self, *_, **__):
            return ""
        def iter(self):
            return []
    blank = _blank()
    def safe(n: Optional[LexborNode]):
        return n or blank
    parser = LexborHTMLParser(r.text)
    hotels = []
    # Use div.uaTTDe for hotel cards
    hotel_cards = parser.css('div.uaTTDe')
    for idx, card in enumerate(hotel_cards):
        # --- NAME EXTRACTION ---
        name = None
        name_elem = card.css_first('h2.BgYkof')
        if name_elem:
            name = name_elem.text(strip=True)
        # --- RATING EXTRACTION ---
        rating = None
        rating_elem = card.css_first('span.KFi5wf.lA0BZ')
        if rating_elem:
            rating_text = rating_elem.text(strip=True)
            try:
                rating = float(rating_text)
            except Exception:
                pass
        else:
            rating_elem = card.css_first('span[aria-label*="out of 5 stars"]')
            if rating_elem:
                aria_label = rating_elem.attributes.get('aria-label', '')
                import re
                m = re.search(r'([0-9.]+) out of 5', aria_label)
                if m:
                    rating = float(m.group(1))
        # --- AMENITIES EXTRACTION ---
        amenities = []
        amenity_selectors = [
            'span.LtjZ2d',
            'span[class*="QYEgn"]',
            'span[class*="amenity"]',
            'div[class*="amenity"]',
            'span[class*="feature"]',
            'div[class*="feature"]'
        ]
        for selector in amenity_selectors:
            amenity_elems = card.css(selector)
            if amenity_elems:
                for a in amenity_elems:
                    text = a.text(strip=True)
                    if text and text not in amenities and len(text) > 2:
                        amenities.append(text)
                if amenities:
                    break
        if not amenities:
            card_text = card.text(strip=True)
            import re
            amenity_patterns = [
                r'Amenities for [^:]+: ([^.]+)',
                r'([A-Za-z\s]+(?:\s*\(\$\))?)(?=,|$)',
            ]
            for pattern in amenity_patterns:
                matches = re.findall(pattern, card_text)
                for match in matches:
                    if isinstance(match, str):
                        potential_amenities = [a.strip() for a in match.split(',') if a.strip() and len(a.strip()) > 2]
                        for amenity in potential_amenities:
                            if amenity not in amenities and not amenity.isdigit():
                                amenities.append(amenity)
                    if amenities:
                        break
        # --- URL EXTRACTION ---
        url = None
        link_elem = card.css_first('a[href]')
        if link_elem:
            url = link_elem.attributes.get('href')
            if url and url.startswith('/travel/'):
                url = 'https://www.google.com' + url
        # --- PRICE EXTRACTION ---
        price = None
        import re
        card_text = card.text(strip=True)
        price_matches = re.findall(r'\$([0-9,.]+)', card_text)
        if price_matches:
            try:
                price = float(price_matches[0].replace(',', ''))
            except Exception:
                price = None
        if name and price is not None:
            hotels.append({
                "name": name,
                "price": price,
                "rating": rating,
                "amenities": amenities,
                "url": url,
            })
    if not hotels:
        # Fallback: try to extract any hotel-like data from the HTML
        import re
        price_pattern = r'\$(\d+(?:,\d+)?)'
        prices = re.findall(price_pattern, r.text)
        name_pattern = r'<h2[^>]*>([^<]+)</h2>'
        names = re.findall(name_pattern, r.text)
        potential_names = []
        for line in r.text.split('\n'):
            line = line.strip()
            if len(line) > 10 and len(line) < 100 and not line.startswith('<') and not line.startswith('$'):
                potential_names.append(line)
        for i, price_str in enumerate(prices[:10]):
            try:
                price = float(price_str.replace(',', ''))
                name = f"Hotel {i+1}"
                if i < len(names):
                    name = names[i].strip()
                elif i < len(potential_names):
                    name = potential_names[i]
                hotels.append({
                    "name": name,
                    "price": price,
                    "rating": None,
                    "amenities": [],
                    "url": None,
                })
            except:
                continue
    if not hotels:
        raise RuntimeError("No hotels found:\n{}".format(r.text_markdown))
    if sort_by == "price":
        hotels.sort(key=lambda h: h["price"], reverse=True)
    elif sort_by == "rating":
        hotels.sort(key=lambda h: h["rating"] or 0, reverse=True)
    else:
        def value_ratio(h):
            if h["rating"] and h["price"] and h["price"] > 0:
                return h["rating"] / h["price"]
            return 0
        hotels.sort(key=value_ratio, reverse=True)
    if limit:
        hotels = hotels[:limit]
    lowest_price = min((h["price"] for h in hotels if h["price"] > 0), default=None)
    return Result(
        hotels=[Hotel(**hotel) for hotel in hotels],
        lowest_price=lowest_price,
        current_price=lowest_price
    ) 